/*
** Управление синтезатором частоты на AD9850
** Для управления частотой генератора используется энкодер
** Вывод информации на ЖК дисплей
*/

#include <LiquidCrystal.h>

#define pin_A 2  // pin 2 - поворот энкодера
#define pin_B 12 // pin 12 - направление вращения энкодера
#define pin_sw 3 // pin 3 - кнопка энкодера

#define S_IN 3 // аналоговый вход 3 для чтения показаний S метра

#define RESET 8 // pin 8 - вход сброса модуля AD9850 (RST)
#define FQ_UD 9  // pin 9 - вход обновления частоты модуля AD9850 (FQ)
#define W_CLK 10  // pin 10 - стробирующий вход модуля AD9850 (CLK)
#define DATA 11  // pin 11 - вход последовательных данных модуля AD9850 (D7)

#define pulseHigh(pin) {digitalWrite(pin, HIGH); digitalWrite(pin, LOW); }

#define band_select 13 // pin 13 - вход переключателя диапазонов
#define bands_number 8 // количество диапазонов

const float band[bands_number*2] = {
                                    0900.0, 1005.0,// диапазон №0 (160 м) частота гетеродина = частота сигнала/2
                                    1400.0, 2200.0,// диапазон №1 (80 м, +75, +90, +Free) частота гетеродина = частота сигнала/2 и т.д...
                                    2900.0, 4200.0,//40m, +41, +49
                                    6750.0, 7350.0,//20m, +22
                                    8740.0, 9378.0,//17m, +16
                                    10495.0, 10775.0,//15m
                                    12440.0, 12500.0,//12m
                                    13995.0, 14855.0//10m
                                   };
byte current_band;      // номер текущего диапазона
double Frequency;       // частота генератора
unsigned long time_s;
float change_step = 50;   // шаг изменения частоты гетеродина грубо - 50 кГц (соостветствует изм. сигнала 100 кГц)
byte tuning_flag = 0;
volatile byte flag_interrupt_turn = 0;
volatile byte flag_interrupt_button = 0;

LiquidCrystal lcd(0, 1, 4, 5, 6, 7);
 
void setup()
{  
  pinMode(pin_A, INPUT);// для чтения с выхода А энкодера
  digitalWrite(pin_A, HIGH);// включаем подтягивающий резистор
  pinMode(pin_B, INPUT);// для чтения с выхода В энкодера
  digitalWrite(pin_B, HIGH);// включаем подтягивающий резистор
  pinMode(pin_sw, INPUT);// для чтения с кнопки энкодера
  digitalWrite(pin_sw, HIGH);// включаем подтягивающий резистор
  pinMode(FQ_UD, OUTPUT);// для выдачи на модуль AD9850
  pinMode(W_CLK, OUTPUT);// для выдачи на модуль AD9850
  pinMode(DATA, OUTPUT);// для выдачи на модуль AD9850
  pinMode(RESET, OUTPUT);// для выдачи на модуль AD9850
  pinMode(band_select, INPUT);// для чтения с переключателя диапазонов
  digitalWrite(band_select, HIGH);// включаем подтягивающий резистор
  pulseHigh(RESET);
  pulseHigh(W_CLK);
  pulseHigh(FQ_UD);  // разрешение последовательного режима AD9850
  lcd.begin(16, 2);
  lcd.setCursor(0, 0);
  lcd.print("DDSynth ver.6.1");
  delay(2000);
  current_band = define_band();
  Frequency = band[current_band*2] + (band[current_band*2+1]-band[current_band*2])/2;// середина текущего диапазона
  sendFrequency(Frequency*1000);// начальная установка частоты AD9850 в Гц
  lcd_print_frequency();
  attachInterrupt(0, enc_turn, FALLING); //выход А энкодера подключен к пин 2 (int 0). Ожидаем срез импульса
  attachInterrupt(1, enc_button, FALLING); //выход кнопки энкодера подключен к пин 3 (int 1). Ожидаем срез импульса
  time_s = millis();
} 
 
void loop()
{
 if(flag_interrupt_turn)
   check_flag_turn();
 if(flag_interrupt_button)
   check_flag_button();
 if (bitRead(current_band, 0) != bitRead(digitalRead(band_select), 0))//Если переключился диапазон
   {
    delay(100);
    current_band = define_band();
    Frequency = band[current_band*2] + (band[current_band*2+1]-band[current_band*2])/2;// середина текущего диапазона
    sendFrequency(Frequency*1000);// установка частоты AD9850 в Гц
    lcd_print_frequency();
   }
   
 if ((millis() - time_s) > 500)//раз в полсекунды проверяем показания S метра
   {
   float Smeter = 10.0*(1.0-(float(analogRead(S_IN)) - 470.0)/(797.0-470.0))-2.0;//напряжение АРУ на входе S метра меняется от 2.4
   lcd.setCursor(13, 0);                                           //до 4 вольт обратно пропорционально уровню сигнала. "-2" - собственные шумы
   lcd.print("S");
   if (Smeter < 0)
      Smeter = 0;
   if (Smeter > 9.5)   
      lcd.print("9+");
   else
      {
       lcd.print(int(Smeter));
       lcd.print(" ");
      }
   time_s = millis();
   }
}

void enc_turn()//обработчик прерываний от вращения энкодера
{
  noInterrupts();//запрет прерываний во избежание дребезга
  flag_interrupt_turn = 1;
  return;
}

void enc_button()//обработчик прерываний от кнопки энкодера
{
  noInterrupts();//запрет прерываний во избежание дребезга
  flag_interrupt_button = 1;
  return;
}

void check_flag_turn()//проверка флага вращения энкодера
{
    delay(2);//задержка перед обработкой прерывания для избежания дребезга
    if(digitalRead(pin_A))//если выход А высокий, значит прерывание ложное
      {
        flag_interrupt_turn = 0;
        interrupts(); //разрешение прерываний
        return;
      }
    
    if(digitalRead(pin_B))
      {
        // выход В в сост. 1, значит, вращение по часовой стрелке
        // увеличиваем не более чем до верхней границы диапазона
        if(Frequency + change_step <= band[current_band*2+1]) Frequency += change_step;
          else Frequency = band[current_band*2+1] ;              
      }   
    else 
      {
        // выход В в 0 сост., значит, вращение против часовой стрелки     
        // уменьшаем, но не ниже нижней границы диапазона
        if(Frequency - change_step >= band[current_band*2]) Frequency -= change_step; 
          else Frequency = band[current_band*2] ;               
      }
    sendFrequency(Frequency*1000);//Установка новой частоты AD9850
    lcd_print_frequency();
    flag_interrupt_turn = 0;
    interrupts(); //разрешение прерываний
    return;
}

void check_flag_button()//проверка флага кнопки энкодера
{
    delay(2);//задержка перед обработкой прерывания для избежания дребезга
    if(digitalRead(pin_sw))//если выход кнопки высокий, значит прерывание ложное
      {
        flag_interrupt_button = 0;
        interrupts(); //разрешение прерываний
        return;
      }
    
    lcd.setCursor(11, 0);
    if (++tuning_flag > 2)
      {
        lcd.print(" ");//убрать индикацию точной настройки
        tuning_flag = 0;
      }
    
    switch(tuning_flag)
    {
      case 0:
        change_step = 50;
        break;  
      case 1:
        change_step = 0.5;
        lcd.print("Q");//индикация режима точной настройки 1
        break;
      case 2:
        change_step = 0.005;
        lcd.print("T");//индикация режима точной настройки 2    
    }
    flag_interrupt_button = 0;
    interrupts(); //разрешение прерываний
    return;
}

void lcd_print_frequency()//вывод частоты на ЖКИ
{
int i, j;
float percent;

lcd.clear();
lcd.print(Frequency*2, 2);// частота сигнала = удвоенная частота гетеродина
lcd.print("kHz");

switch(tuning_flag)
    {
      case 0:
        break;  
      case 1:
        lcd.setCursor(11, 0);
        lcd.print("Q");//индикация режима точной настройки 1
        break;
      case 2:
        lcd.setCursor(11, 0);
        lcd.print("T");//индикация режима точной настройки 2
    }

lcd.setCursor(13, 0);
lcd.print("S?");

percent = 100*(Frequency - band[current_band*2])/(band[current_band*2+1] - band[current_band*2]);

for (i = 0; i < 17; i++)
  {
   if(percent<6)
    break;
   for(j = 0; j < int(percent/6.25); j++)
    {
     lcd.setCursor(j, 1);
     lcd.write(255);
    }
  }
return;
}

void tfr_byte(byte data)// последовательная передача байта в модуль AD9850 на вход DATA
{
  for (int i=0; i<8; i++, data>>=1)
  {
    digitalWrite(DATA, data & 0x01);
    pulseHigh(W_CLK);   //строб каждого бита
  }
}
 
 
void sendFrequency(double frequency)// расчет и передача управляющего 32 битного числа "frequency tuning word"
                                    //для задания частоты: frequency = <sys clock> * <frequency tuning word>/2^32
{
  unsigned long freq = frequency * 1.000016 * 4294967295/125000000;  // расчет числа, 1.000016 - поправка для
  for (int b=0; b<4; b++, freq>>=8)                                 // конкретного экземпляра платы
   {
    tfr_byte(freq & 0xFF);//передача числа в AD9850
   }
  tfr_byte(0x000);   // завершающий байт, все 0
  pulseHigh(FQ_UD);  // строб новой частоты
}

byte define_band(void)//Считывание номера текущего диапазона
{
 byte digit1;
 byte digit2;

  if(analogRead(4)<512)
    digit1=0;
  else digit1=2;
  if(analogRead(5)<512)
    digit2=0;
  else digit2=4;
return(digitalRead(band_select)+digit1+digit2);
}
